from __future__ import annotations

from collections import Counter
from typing import Dict, Iterable, List, Tuple

from .grammar import Grammar


class RePair:
    """RePair-style grammar inducer (recursive pair substitution).

    Algorithm (baseline implementation):
      1) Repeatedly find the most frequent adjacent pair (digram) in the sequence.
      2) If its frequency < 2, stop.
      3) Introduce a new non-terminal rule LHS -> (a, b) and replace all
         non-overlapping occurrences of (a, b) with LHS.
      4) After convergence, enforce rule utility: inline rules used <= 1 time.
      5) Compute rule usage counts.

    This is a pragmatic, readable version to bootstrap further optimization.
    """

    def __init__(self, prefix: str = "R") -> None:
        self.prefix = prefix

    def compress(self, tokens: List[str]) -> Tuple[List[str], Grammar]:
        sequence = list(tokens)
        grammar = Grammar()
        rule_id = 1

        while True:
            counts = self._digram_counts(sequence)
            if not counts:
                break
            digram, freq = max(counts.items(), key=lambda kv: kv[1])
            if freq < 2:
                break

            lhs = f"{self.prefix}{rule_id}"
            grammar.add_rule(lhs, [digram[0], digram[1]])
            rule_id += 1
            sequence = self._replace_all(sequence, digram, lhs)

        # Rule utility: inline rules used <= 1 time
        self._inline_singletons(sequence, grammar)

        # Compute usage counts for rules (in compressed seq + nested in RHS)
        self._update_rule_usage(sequence, grammar)

        return sequence, grammar

    def compress_trace(self, tokens: List[str]) -> List[Tuple[List[str], Grammar]]:
        """Run RePair and return snapshots after each replacement step.

        Note: snapshots are recorded before final singleton inlining; usage
        counts are updated for each snapshot to allow entropy/MDL measurement.
        The final step applies inlining and usage update and is also included.
        """
        sequence = list(tokens)
        grammar = Grammar()
        rule_id = 1
        snapshots: List[Tuple[List[str], Grammar]] = []

        while True:
            counts = self._digram_counts(sequence)
            if not counts:
                break
            digram, freq = max(counts.items(), key=lambda kv: kv[1])
            if freq < 2:
                break

            lhs = f"{self.prefix}{rule_id}"
            grammar.add_rule(lhs, [digram[0], digram[1]])
            rule_id += 1
            sequence = self._replace_all(sequence, digram, lhs)

            # Update usage for snapshot
            self._update_rule_usage(sequence, grammar)
            snapshots.append((list(sequence), grammar.clone()))

        # Finalize with inlining and update usage, include final snapshot
        self._inline_singletons(sequence, grammar)
        self._update_rule_usage(sequence, grammar)
        snapshots.append((list(sequence), grammar.clone()))
        return snapshots

    def reconstruct(self, compressed: List[str], grammar: Grammar) -> List[str]:
        """Expand compressed tokens using grammar rules until terminals only."""
        def expand_symbol(sym: str) -> List[str]:
            if sym in grammar.rules:
                rhs = [s for s in grammar.as_tuples()[sym]]
                out: List[str] = []
                for t in rhs:
                    out.extend(expand_symbol(t))
                return out
            return [sym]

        out: List[str] = []
        for t in compressed:
            out.extend(expand_symbol(t))
        return out

    @staticmethod
    def _digram_counts(sequence: List[str]) -> Dict[Tuple[str, str], int]:
        return Counter(zip(sequence, sequence[1:]))

    @staticmethod
    def _replace_all(sequence: List[str], digram: Tuple[str, str], lhs: str) -> List[str]:
        out: List[str] = []
        i = 0
        n = len(sequence)
        while i < n:
            if i < n - 1 and sequence[i] == digram[0] and sequence[i + 1] == digram[1]:
                out.append(lhs)
                i += 2
            else:
                out.append(sequence[i])
                i += 1
        return out

    def _inline_singletons(self, sequence: List[str], grammar: Grammar) -> None:
        # Repeat until no change: inline rules that appear <= 1 time across sequence and rules
        changed = True
        while changed and grammar.rules:
            changed = False
            usage = self._collect_symbol_usage(sequence, grammar)
            for lhs in list(grammar.rules.keys()):
                if usage.get(lhs, 0) <= 1:
                    rhs = list(grammar.as_tuples()[lhs])
                    # Inline once in sequence if present
                    if lhs in sequence:
                        idx = sequence.index(lhs)
                        sequence[idx:idx + 1] = rhs
                        changed = True
                    # Inline inside other rules
                    for r in grammar.rules.values():
                        if r.lhs.value == lhs:
                            continue
                        new_rhs: List[str] = []
                        for sym in [s.value for s in r.rhs]:
                            if sym == lhs:
                                new_rhs.extend(rhs)
                                changed = True
                            else:
                                new_rhs.append(sym)
                        r.rhs = [grammar._get_symbol(v) for v in new_rhs]
                    # Remove the rule
                    if lhs in grammar.rules:
                        del grammar.rules[lhs]

    def _collect_symbol_usage(self, sequence: List[str], grammar: Grammar) -> Dict[str, int]:
        usage: Dict[str, int] = Counter(sequence)
        for lhs, rule in grammar.rules.items():
            for sym in rule.rhs:
                usage[sym.value] = usage.get(sym.value, 0) + 1
        return usage

    def _update_rule_usage(self, sequence: List[str], grammar: Grammar) -> None:
        usage = self._collect_symbol_usage(sequence, grammar)
        # Count only uses of each LHS (not internal symbol frequency)
        for lhs, rule in grammar.rules.items():
            rule.frequency = usage.get(lhs, 0)
        # Optionally compute probabilities as normalized frequencies across rules
        total = sum(r.frequency for r in grammar.rules.values()) or 1
        for r in grammar.rules.values():
            r.probability = r.frequency / total



from __future__ import annotations

import os
from typing import List, Optional


def _openai_refine(prompt: str) -> Optional[str]:
    try:
        import openai  # type: ignore
    except Exception:
        return None
    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        return None
    try:
        client = openai.OpenAI(api_key=api_key)  # newer SDK
        resp = client.chat.completions.create(
            model=os.getenv("UFR_OPENAI_MODEL", "gpt-4o-mini"),
            messages=[{"role": "user", "content": prompt}],
            temperature=float(os.getenv("UFR_OPENAI_T", "0.2")),
        )
        return resp.choices[0].message.content  # type: ignore
    except Exception:
        return None


def _anthropic_refine(prompt: str) -> Optional[str]:
    try:
        import anthropic  # type: ignore
    except Exception:
        return None
    api_key = os.getenv("ANTHROPIC_API_KEY")
    if not api_key:
        return None
    try:
        client = anthropic.Anthropic(api_key=api_key)
        msg = client.messages.create(
            model=os.getenv("UFR_ANTHROPIC_MODEL", "claude-3-5-sonnet-20240620"),
            max_tokens=2000,
            temperature=float(os.getenv("UFR_ANTHROPIC_T", "0.2")),
            messages=[{"role": "user", "content": prompt}],
        )
        # anthropic SDK returns content blocks
        parts = getattr(msg, "content", [])
        if parts and hasattr(parts[0], "text"):
            return parts[0].text  # type: ignore
        return None
    except Exception:
        return None


def refine_component(name: str, variants: List[str], props: List[str], base_tsx: str, provider: Optional[str] = None) -> Optional[str]:
    """Optionally refine a generated component via an LLM.

    Provider is 'openai' or 'anthropic'; reads API keys from env.
    Returns refined TSX string or None if unavailable/failure.
    """
    prompt = f"""
You are assisting with a React component refactor guided by discovered UI grammar patterns.
Please improve the following TypeScript React component skeleton while preserving prop and variant contracts.

Component name: {name}
Variants: {variants}
Props: {props}

Goals:
- Keep the API stable; do not remove props.
- If helpful, add better types for common props (e.g., variant as union, booleans, strings).
- Ensure accessibility basics (labels, aria- attributes where reasonable) and good defaults.
- Keep the implementation simple; stubs are fine but avoid dead code.

Provided skeleton:
```
{base_tsx}
```

Return only the improved TSX file content.
"""
    if provider == "openai":
        return _openai_refine(prompt)
    if provider == "anthropic":
        return _anthropic_refine(prompt)
    return None


def generate_text(prompt: str, provider: Optional[str] = None) -> Optional[str]:
    """Generic text generation for summaries via the selected provider."""
    if provider == "openai":
        return _openai_refine(prompt)
    if provider == "anthropic":
        return _anthropic_refine(prompt)
    return None



"""Emergence Engine core package."""

from .symbols import Symbol
from .grammar import ProductionRule, Grammar
from .sequitur import Sequitur
from .repair import RePair
from .mdl import MDLScorer
from .engine import EmergenceEngine

__all__ = [
    "Symbol",
    "ProductionRule",
    "Grammar",
    "Sequitur",
    "RePair",
    "MDLScorer",
    "EmergenceEngine",
]



from __future__ import annotations

from collections import Counter
from typing import Dict, List, Tuple

from .grammar import Grammar


class Sequitur:
    """A simple SEQUITUR-inspired digram compressor.

    This is a pragmatic baseline: find most frequent digrams in the current
    sequence, introduce non-terminals, and replace all non-overlapping
    occurrences until no repeated digrams remain.
    """

    def __init__(self) -> None:
        self.rule_prefix = "R"

    def compress(self, tokens: List[str]) -> Tuple[List[str], Grammar]:
        sequence = list(tokens)
        grammar = Grammar()
        rule_counter = 1

        while True:
            counts = self._digram_counts(sequence)
            if not counts:
                break

            # Find the most frequent digram (freq >= 2)
            digram, freq = max(counts.items(), key=lambda kv: kv[1])
            if freq < 2:
                break

            # Introduce or reuse a rule for this digram
            lhs = f"{self.rule_prefix}{rule_counter}"
            grammar.add_rule(lhs, [digram[0], digram[1]])
            rule_counter += 1

            # Replace all non-overlapping occurrences
            sequence = self._replace_digram(sequence, digram, lhs)

        # Simple rule utility: inline rules that occur only once
        changed = True
        while changed:
            changed = False
            for lhs, rhs in list(grammar.as_tuples().items()):
                occurrences = sequence.count(lhs)
                if occurrences <= 1:
                    # Inline once and drop the rule
                    if occurrences == 1:
                        sequence = self._inline_once(sequence, lhs, list(rhs))
                    if lhs in grammar.rules:
                        del grammar.rules[lhs]
                    changed = True

        return sequence, grammar

    @staticmethod
    def _digram_counts(sequence: List[str]) -> Dict[Tuple[str, str], int]:
        counts: Dict[Tuple[str, str], int] = {}
        for i in range(len(sequence) - 1):
            digram = (sequence[i], sequence[i + 1])
            counts[digram] = counts.get(digram, 0) + 1
        return counts

    @staticmethod
    def _replace_digram(sequence: List[str], digram: Tuple[str, str], symbol: str) -> List[str]:
        out: List[str] = []
        i = 0
        n = len(sequence)
        while i < n:
            if i < n - 1 and sequence[i] == digram[0] and sequence[i + 1] == digram[1]:
                out.append(symbol)
                i += 2
            else:
                out.append(sequence[i])
                i += 1
        return out

    @staticmethod
    def _inline_once(sequence: List[str], lhs: str, rhs: List[str]) -> List[str]:
        out: List[str] = []
        inlined = False
        for tok in sequence:
            if not inlined and tok == lhs:
                out.extend(rhs)
                inlined = True
            else:
                out.append(tok)
        return out




from __future__ import annotations

from typing import Dict, List

from .grammar import Grammar


def grammar_to_dot(grammar: Grammar) -> str:
    """Render a simple GraphViz DOT for the grammar rules.

    Each rule LHS -> RHS is represented as a node for LHS and edges to each RHS symbol.
    Terminals are boxed; non-terminals are ellipses.
    """
    lines: List[str] = ["digraph Grammar {", "  rankdir=LR;"]
    # Declare nodes
    nts = set(grammar.non_terminals.keys())
    ts = set(grammar.terminals.keys())
    for t in ts:
        lines.append(f'  "{t}" [shape=box, style=filled, fillcolor=lightgray];')
    for nt in nts:
        lines.append(f'  "{nt}" [shape=ellipse];')
    # Edges per rule
    for lhs, rule in grammar.rules.items():
        for sym in rule.rhs:
            lines.append(f'  "{lhs}" -> "{sym.value}";')
    lines.append("}")
    return "\n".join(lines)


def trajectory_to_plotly_json(entropies: List[float], mdl_traj: List[Dict[str, float]]) -> Dict:
    """Return a minimal Plotly-friendly JSON payload with entropy and MDL curves."""
    x = list(range(len(entropies)))
    total = [step["total"] for step in mdl_traj]
    data = [
        {"type": "scatter", "name": "Entropy", "x": x, "y": entropies, "yaxis": "y1"},
        {"type": "scatter", "name": "MDL Total", "x": x, "y": total, "yaxis": "y2"},
    ]
    layout = {
        "title": "Entropy and MDL Trajectory",
        "xaxis": {"title": "Step"},
        "yaxis": {"title": "Entropy"},
        "yaxis2": {"title": "MDL", "overlaying": "y", "side": "right"},
    }
    return {"data": data, "layout": layout}




from __future__ import annotations

from dataclasses import dataclass
from math import log2
from typing import Dict, List, Tuple, Optional

from .grammar import Grammar


def compute_entropy(grammar: Grammar) -> float:
    """Shannon entropy (bits) over rule usage frequencies.

    If there are no rules or zero total frequency, returns 0.0.
    """
    total = sum(r.frequency for r in grammar.rules.values())
    if total <= 0:
        return 0.0
    H = 0.0
    for r in grammar.rules.values():
        if r.frequency <= 0:
            continue
        p = r.frequency / total
        H -= p * log2(p)
    return H


@dataclass
class EmergenceEvent:
    index: int
    magnitude: float
    kind: str  # 'emergence' or 'dissolution'
    entropy_before: float
    entropy_after: float
    rules_added: List[str]


class EmergenceDetector:
    def __init__(
        self,
        threshold: float = 0.25,
        window: int = 1,
        *,
        preset: Optional[str] = None,
        mode: str = "static",  # 'static' | 'adaptive'
        k: float = 3.0,
        percentile: float = 0.9,
        min_persistence: int = 2,
        hysteresis: float = 0.1,
        min_gap: int = 2,
    ) -> None:
        """Initialize detector.

        - threshold: normalized curvature threshold (abs(d2)/max_entropy)
        - window: derivative window (reserved)
        - preset: 'sensitive'|'balanced'|'strict' sets threshold if provided
        - mode: 'static' uses given/preset threshold; 'adaptive' computes from data
        - k: tuning parameter for robust/zscore modes (currently used for robust)
        - percentile: for percentile-based thresholding (not used yet)
        """
        self.threshold = threshold
        self.window = window
        self.preset = preset
        self.mode = mode
        self.k = k
        self.percentile = percentile
        self.min_persistence = max(1, int(min_persistence))
        self.hysteresis = max(0.0, float(hysteresis))
        self.min_gap = max(0, int(min_gap))

        if preset is not None:
            self.threshold = self._preset_threshold(preset)

    @staticmethod
    def _preset_threshold(name: str) -> float:
        name = name.lower()
        if name == "sensitive":
            return 0.15
        if name == "balanced":
            return 0.25
        if name == "strict":
            return 0.40
        return 0.25

    @staticmethod
    def _curvatures(entropies: List[float]) -> List[float]:
        if len(entropies) < 3:
            return []
        return [entropies[i + 1] - 2 * entropies[i] + entropies[i - 1] for i in range(1, len(entropies) - 1)]

    def threshold_from_entropies(self, entropies: List[float]) -> float:
        """Compute threshold for 'adaptive' mode using robust stats on curvature.

        thr = median(|d2|/max_entropy) + k * MAD(|d2|/max_entropy)
        Fallback to default threshold if insufficient points.
        """
        curv = self._curvatures(entropies)
        if len(curv) == 0:
            return self.threshold
        max_entropy = max(entropies) or 1.0
        values = [abs(c) / max_entropy for c in curv]
        med = _median(values)
        mad = _median([abs(v - med) for v in values])
        return med + self.k * mad

    def detect(self, grammars: List[Grammar]) -> Dict[str, List]:
        entropies = [compute_entropy(g) for g in grammars]
        events: List[EmergenceEvent] = []
        if len(entropies) < 3:
            return {"entropies": entropies, "events": events}

        thr = self.threshold if self.mode != "adaptive" else self.threshold_from_entropies(entropies)
        max_entropy = max(entropies) or 1.0

        # Persistence/hysteresis gating
        run = 0
        active = False
        last_event_idx = -10**9

        for i in range(1, len(entropies) - 1):
            d2 = entropies[i + 1] - 2 * entropies[i] + entropies[i - 1]
            norm = abs(d2) / max_entropy
            if norm >= thr:
                run += 1
            else:
                run = 0

            if not active and run >= self.min_persistence and (i - last_event_idx) >= self.min_gap:
                kind = "emergence" if d2 < 0 else "dissolution"
                events.append(
                    EmergenceEvent(
                        index=i,
                        magnitude=d2,
                        kind=kind,
                        entropy_before=entropies[i - 1],
                        entropy_after=entropies[i + 1],
                        rules_added=[],
                    )
                )
                active = True
                last_event_idx = i

            if active and norm <= max(0.0, thr - self.hysteresis):
                active = False

        return {"entropies": entropies, "events": events}


def _median(values: List[float]) -> float:
    n = len(values)
    if n == 0:
        return 0.0
    vs = sorted(values)
    mid = n // 2
    if n % 2 == 1:
        return vs[mid]
    return 0.5 * (vs[mid - 1] + vs[mid])



from __future__ import annotations

import json
import shutil
import subprocess
from pathlib import Path
from typing import Dict, List, Tuple


def has_node_tokenizer() -> bool:
    node = shutil.which("node")
    script = Path("tools/js-tokenizer/index.js")
    return bool(node and script.exists())


def tokenize_project_ast(root_dir: str | Path) -> Tuple[List[str], Dict[str, int]]:
    root = Path(root_dir)
    script = Path("tools/js-tokenizer/index.js").resolve()
    proc = subprocess.run(
        ["node", str(script), str(root.resolve())],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        text=True,
    )
    if proc.returncode != 0:
        raise RuntimeError(f"Node tokenizer failed: {proc.stderr.strip()}")
    data = json.loads(proc.stdout)
    tokens = list(data.get("tokens", []))
    counts = {
        "files": int(data.get("files", 0)),
        "tags": int(data.get("counts", {}).get("tags", 0)),
        "props": int(data.get("counts", {}).get("props", 0)),
        "components": int(data.get("counts", {}).get("components", 0)),
    }
    return tokens, counts




from dataclasses import dataclass


@dataclass(frozen=True)
class Symbol:
    """Simple symbol wrapper (type, value).

    For now, we mostly use string values; `kind` can distinguish terminals vs non-terminals.
    """

    value: str
    kind: str = "terminal"  # or "nonterminal"

    def __str__(self) -> str:  # pragma: no cover - trivial
        return self.value




from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path

from .engine import EmergenceEngine


def parse_args(argv: list[str]) -> argparse.Namespace:
    p = argparse.ArgumentParser(prog="emergence", description="Emergence Engine CLI")
    p.add_argument("path", nargs="?", help="Path to input file (default: stdin)")
    p.add_argument("--chars", action="store_true", help="Character-level tokenization")
    p.add_argument("--pretty", action="store_true", help="Pretty-print JSON output")
    p.add_argument("--emergence", action="store_true", help="Compute entropy trajectory and events")
    p.add_argument("--threshold", type=float, default=0.25, help="Emergence normalized curvature threshold")
    p.add_argument("--window", type=int, default=1, help="Emergence detection window (reserved)")
    p.add_argument("--emergence-preset", choices=["sensitive", "balanced", "strict"], help="Preset for emergence detection")
    p.add_argument("--emergence-mode", choices=["static", "adaptive"], default="static", help="Static uses given/preset threshold; adaptive computes from data")
    p.add_argument("--k", type=float, default=3.0, help="Adaptive robust threshold multiplier (MAD)")
    p.add_argument("--percentile", type=float, default=0.9, help="Adaptive percentile (reserved)")
    p.add_argument("--export-graphviz", help="Write final grammar GraphViz DOT to file")
    p.add_argument("--export-plotly", help="Write trajectory Plotly JSON to file")
    p.add_argument("--no-trajectory", action="store_true", help="With --emergence, omit mdl_trajectory from stdout JSON")
    p.add_argument("--summary", action="store_true", help="Print concise summary fields only")
    p.add_argument("--sliding-window", type=int, help="Analyze with sliding windows of this token length")
    p.add_argument("--sliding-step", type=int, default=0, help="Step between windows (default: window/2)")
    p.add_argument("--min-persistence", type=int, default=2, help="Emergence: min consecutive steps above threshold")
    p.add_argument("--hysteresis", type=float, default=0.1, help="Emergence: hysteresis margin to end events")
    p.add_argument("--min-gap", type=int, default=2, help="Emergence: minimum gap between events")
    return p.parse_args(argv)


def main(argv: list[str] | None = None) -> int:
    argv = list(sys.argv[1:] if argv is None else argv)
    args = parse_args(argv)

    text: str
    if args.path:
        text = Path(args.path).read_text(encoding="utf-8")
    else:
        text = sys.stdin.read()

    engine = EmergenceEngine()
    # Sliding window mode
    if args.sliding_window:
        step = args.sliding_step if args.sliding_step > 0 else max(1, args.sliding_window // 2)
        result = engine.process_sliding(
            text,
            window_size=args.sliding_window,
            step=step,
            threshold=args.threshold,
            preset=args.emergence_preset,
            mode=args.emergence_mode,
            k=args.k,
            percentile=args.percentile,
            chars=args.chars,
            min_persistence=args.min_persistence,
            hysteresis=args.hysteresis,
            min_gap=args.min_gap,
        )
    else:
        result = engine.process(
            text,
            emergence=args.emergence,
            threshold=args.threshold,
            window=args.window,
            preset=args.emergence_preset,
            mode=args.emergence_mode,
            k=args.k,
            percentile=args.percentile,
            chars=args.chars,
            min_persistence=args.min_persistence,
            hysteresis=args.hysteresis,
            min_gap=args.min_gap,
        )

    if args.summary:
        payload = {
            "compression_ratio": result.compression_ratio,
            "mdl_score": result.mdl_score,
            "mdl_grammar_cost": result.mdl_grammar_cost,
            "mdl_data_cost": result.mdl_data_cost,
            "naive_mdl": result.naive_mdl,
            "coverage": result.coverage,
            "valid_lossless": result.valid_lossless,
            # Include compact emergence summary when requested
        }
    else:
        payload = {
            "compressed": result.compressed,
            "rules": result.rules,
            "mdl_score": result.mdl_score,
            "mdl_grammar_cost": result.mdl_grammar_cost,
            "mdl_data_cost": result.mdl_data_cost,
            "naive_mdl": result.naive_mdl,
            "compression_ratio": result.compression_ratio,
            "coverage": result.coverage,
            "valid_lossless": result.valid_lossless,
        }

    if args.emergence and not args.sliding_window:
        if args.summary:
            payload["entropies"] = {
                "count": len(result.entropies or []),
                "first": (result.entropies or [None])[0],
                "last": (result.entropies or [None])[-1],
            }
            payload["events_summary"] = {
                "count": len(result.events or []),
                "kinds": sorted({e.get("kind") for e in (result.events or [])}),
                "indices": [e.get("index") for e in (result.events or [])][:10],
            }
        else:
            payload["entropies"] = result.entropies
            payload["events"] = result.events
            if not args.no_trajectory:
                # Also include MDL trajectory unless suppressed
                payload["mdl_trajectory"] = getattr(result, "mdl_trajectory", None)
    # Sliding window outputs
    if args.sliding_window:
        if args.summary:
            payload["windows"] = {
                "count": len(result.windows_mdl or []),
                "first_mdl": (result.windows_mdl or [None])[0],
                "last_mdl": (result.windows_mdl or [None])[-1],
            }
            payload["window_events_summary"] = {
                "count": len(result.window_events or []),
                "kinds": sorted({e.get("kind") for e in (result.window_events or [])}),
                "indices": [e.get("index") for e in (result.window_events or [])][:10],
            }
        else:
            payload["windows_entropies"] = result.windows_entropies
            payload["windows_mdl"] = result.windows_mdl
            payload["window_events"] = result.window_events

        # Optional exports
        if args.export_graphviz:
            from .viz import grammar_to_dot
            try:
                from .grammar import Grammar
                g = Grammar()
                for lhs, rhs in result.rules.items():
                    g.add_rule(lhs, list(rhs))
                Path(args.export_graphviz).write_text(grammar_to_dot(g), encoding="utf-8")
            except Exception as e:
                print(f"Warning: failed to export GraphViz: {e}", file=sys.stderr)
        if args.export_plotly and payload.get("entropies") and payload.get("mdl_trajectory"):
            from .viz import trajectory_to_plotly_json
            try:
                import json as _json
                Path(args.export_plotly).write_text(
                    _json.dumps(trajectory_to_plotly_json(payload["entropies"], payload["mdl_trajectory"]), indent=2),
                    encoding="utf-8",
                )
            except Exception as e:
                print(f"Warning: failed to export Plotly JSON: {e}", file=sys.stderr)

    if args.pretty:
        print(json.dumps(payload, indent=2))
    else:
        print(json.dumps(payload))

    return 0


if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())



from __future__ import annotations

from pathlib import Path
from typing import List, Dict, Optional
from .llm import refine_component
from string import Template


TEMPLATE_TSX = Template("""// Auto-generated by UFR-DS (prototype)
import React from 'react';

export type { ReactNode } from 'react';

export interface ${name}Props {
  variant?: ${variant_union};
  ${prop_fields}
  children?: React.ReactNode;
}

export function ${name}({ variant = ${variant_default}, ${prop_params} children }: ${name}Props) {
  // TODO: Replace with real implementation mapped from your design system
  const className = 'ufr-${kebab}' + (variant ? ' variant-' + variant : '');
  return (
    <div className={className}>
      {/* Render children and map props into structure as needed */}
      {children}
    </div>
  );
}
""")


def kebab_case(s: str) -> str:
    out = []
    for c in s:
        if c.isupper():
            if out:
                out.append('-')
            out.append(c.lower())
        else:
            out.append(c)
    return ''.join(out)


def write_component(out_dir: Path, name: str, variant_values: List[str], props: List[str], *, llm_provider: Optional[str] = None) -> Path:
    out_dir.mkdir(parents=True, exist_ok=True)
    file = out_dir / f"{name}.tsx"
    if not variant_values:
        variant_values = ["default"]
    # sanitize and dedupe
    variant_values = [str(v) for v in variant_values if v]
    variant_union = " | ".join(sorted({f"'{v}'" for v in variant_values})) or "'default'"
    variant_default = f"'{variant_values[0]}'"
    # Prepare props
    prop_fields = "\n  ".join(f"{p}?: any;" for p in props)
    if prop_fields:
        prop_fields = prop_fields + "\n  "
    prop_params = ", ".join(props)
    if prop_params:
        prop_params = prop_params + ", "
    content = TEMPLATE_TSX.substitute(
        name=name,
        variant_union=variant_union,
        variant_default=variant_default,
        prop_fields=prop_fields,
        prop_params=prop_params,
        kebab=kebab_case(name),
    )
    # Optional LLM refinement
    refined = refine_component(name, variant_values, props, content, provider=llm_provider) if llm_provider else None
    file.write_text(refined or content, encoding="utf-8")
    return file


def generate_from_suggestions(suggestions: List[dict], out_dir: str | Path, *, llm_provider: Optional[str] = None) -> Dict[str, str]:
    out = {}
    out_path = Path(out_dir)
    for s in suggestions:
        name = s.get("family") or "Component"
        variants = list(s.get("suggested_variant_values") or [])
        props = list(s.get("top_props") or [])
        path = write_component(out_path, name, variants, props, llm_provider=llm_provider)
        out[name] = str(path)
    # Write an index.ts for convenience
    if out:
        index = out_path / "index.ts"
        exports = "\n".join([f"export * from './{k}';" for k in out.keys()])
        index.write_text(exports + "\n", encoding='utf-8')
    return out



from __future__ import annotations

import re
from pathlib import Path
from typing import Iterable, List, Tuple, Dict, Set


JS_EXTS = {".js", ".jsx", ".ts", ".tsx"}


def list_source_files(root: Path) -> List[Path]:
    files: List[Path] = []
    for p in root.rglob("*"):
        if p.is_file() and p.suffix in JS_EXTS:
            files.append(p)
    return files


TAG_OPEN_RE = re.compile(r"<\s*([A-Za-z_][A-Za-z0-9_]*)[\s>/]")
TAG_CLOSE_RE = re.compile(r"</\s*([A-Za-z_][A-Za-z0-9_]*)\s*>")
PROP_RE = re.compile(r"\b([A-Za-z_][A-Za-z0-9_-]*)\s*=\s*")
IMPORT_NAMED_RE = re.compile(r"import\s*\{([^}]*)\}\s*from\s*['\"][^'\"]+['\"];?")
IMPORT_DEFAULT_RE = re.compile(r"import\s+([A-Za-z_][A-Za-z0-9_]*)\s+from\s*['\"][^'\"]+['\"];?")


HTML_TAGS: Set[str] = {
    # Common HTML tags to optionally downweight or skip; we keep them but can filter later
    "div", "span", "p", "a", "img", "ul", "ol", "li", "input", "button", "select", "option",
    "label", "form", "table", "thead", "tbody", "tr", "td", "th", "header", "footer", "main",
    "section", "article", "nav", "aside", "h1", "h2", "h3", "h4", "h5", "h6",
}


def tokenize_file(path: Path) -> List[str]:
    try:
        text = path.read_text(encoding="utf-8", errors="ignore")
    except Exception:
        return []

    tokens: List[str] = []

    # Imports
    for m in IMPORT_DEFAULT_RE.finditer(text):
        name = m.group(1)
        tokens.append(f"IMPORT:{name}")
    for m in IMPORT_NAMED_RE.finditer(text):
        names = [n.strip() for n in m.group(1).split(",") if n.strip()]
        for name in names:
            # Handle aliasing: X as Y
            name = name.split(" as ")[0].strip()
            if name:
                tokens.append(f"IMPORT:{name}")

    # JSX tags
    for m in TAG_OPEN_RE.finditer(text):
        tag = m.group(1)
        tokens.append(f"TAG:{tag}")
    for m in TAG_CLOSE_RE.finditer(text):
        tag = m.group(1)
        tokens.append(f"TAG_CLOSE:{tag}")

    # JSX props
    for m in PROP_RE.finditer(text):
        prop = m.group(1)
        # Filter out likely non-prop assignments in code by simple heuristic: occurs near a '<'
        # Keep all for now; engine can compress noise away
        tokens.append(f"PROP:{prop}")

    return tokens


def tokenize_project(root_dir: str | Path) -> Tuple[List[str], Dict[str, int]]:
    """Tokenize a React/TS project by extracting JSX tags, props, and imports.

    Returns a flat token sequence and basic counts for quick summaries.
    """
    root = Path(root_dir)
    all_tokens: List[str] = []
    counts: Dict[str, int] = {
        "files": 0,
        "tags": 0,
        "components": 0,
        "props": 0,
    }

    for file in list_source_files(root):
        counts["files"] += 1
        toks = tokenize_file(file)
        all_tokens.extend(toks)

    # Summaries
    counts["tags"] = sum(1 for t in all_tokens if t.startswith("TAG:"))
    counts["components"] = sum(1 for t in all_tokens if t.startswith("IMPORT:"))
    counts["props"] = sum(1 for t in all_tokens if t.startswith("PROP:"))
    return all_tokens, counts




from __future__ import annotations

from math import log2, floor
from typing import Dict, List, Tuple

from .grammar import Grammar


def _safe_log2(x: int) -> float:
    return log2(max(2, x))


def elias_gamma_length(n: int) -> int:
    """Bit-length of Elias gamma code for positive integer n (n>=1).

    For n<=0, we map to 1 to keep things defined.
    L(n) = 2*floor(log2(n)) + 1
    """
    if n <= 0:
        n = 1
    return 2 * floor(log2(n)) + 1


class MDLScorer:
    """Explicit two-part MDL with simple universal codes.

    - Grammar cost (lightweight):
      - Sum log2(V) for symbols in each RHS.
      - We intentionally omit explicit costs for number of rules,
        RHS length headers, and rule frequency to avoid over-penalizing
        simple, highly-repetitive cases.
    - Data cost:
      - Encode sequence length via gamma(n)
      - Then n * log2(V) to emit symbol indices

    V = |Σ ∪ N|, vocabulary size (terminals + non-terminals)
    Σ size is provided via `terminals_size` when scoring.
    """

    def score_components(self, grammar: Grammar, compressed: List[str], terminals_size: int) -> Dict[str, float]:
        sigma = max(2, terminals_size)
        V = max(2, sigma + len(grammar.rules))
        sym_cost = _safe_log2(V)

        # Grammar cost: only pay for symbols appearing in RHS across rules
        grammar_cost = 0.0
        for rule in grammar.rules.values():
            rhs_len = len(rule.rhs)
            grammar_cost += rhs_len * sym_cost

        # Data cost
        n = len(compressed)
        data_cost = float(elias_gamma_length(n)) + n * sym_cost

        total = grammar_cost + data_cost
        return {"grammar_cost": grammar_cost, "data_cost": data_cost, "total": total}

    def score(self, grammar: Grammar, compressed: List[str], terminals_size: int | None = None) -> float:
        sigma = terminals_size if terminals_size is not None else max(2, len(grammar.terminals) or 2)
        return self.score_components(grammar, compressed, sigma)["total"]

    def naive_baseline(self, original_tokens: List[str]) -> float:
        sigma = max(2, len(set(original_tokens)) or 2)
        n = len(original_tokens)
        return float(elias_gamma_length(n)) + n * _safe_log2(sigma)

    def compression_ratio(self, original_tokens: List[str], compressed: List[str], grammar: Grammar) -> float:
        sigma = max(2, len(set(original_tokens)) or 2)
        original_size = self.naive_baseline(original_tokens)
        compressed_size = self.score(grammar, compressed, terminals_size=sigma)
        if compressed_size == 0:
            return 1.0
        return original_size / compressed_size

    def delta(self, before: Tuple[List[str], Grammar], after: Tuple[List[str], Grammar], terminals_size: int) -> float:
        b_seq, b_g = before
        a_seq, a_g = after
        return self.score(a_g, a_seq, terminals_size) - self.score(b_g, b_seq, terminals_size)



from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import List

from emergence_engine.engine import EmergenceEngine

from .react_tokenizer import tokenize_project
from .tokenizer_node_bridge import has_node_tokenizer, tokenize_project_ast
from .families import discover_families, family_suggestions
from .engine_suggest import suggestions_from_rules
from .generator import generate_from_suggestions


# Self-contained report handler to avoid NameError during parser construction
def _report_cmd(args: argparse.Namespace) -> int:
    from .report import render_markdown_report, render_llm_summary
    data = json.loads(Path(args.analysis).read_text(encoding='utf-8'))
    md = render_markdown_report(data)
    if getattr(args, 'llm', None):
        summary = render_llm_summary(data, args.llm)
        if summary:
            md += "\n\n## Human Summary (LLM)\n\n" + summary + "\n"
        else:
            md += "\n\n_note: LLM summary unavailable (no API key or provider error)._\n"
    args.out.write_text(md, encoding='utf-8')
    print(f"Wrote report to {args.out}")
    return 0


def cmd_analyze(args: argparse.Namespace) -> int:
    # Input source precedence: --tokens JSON > AST tokenizer > regex tokenizer
    tokens: list[str]
    counts: dict[str, int]
    if getattr(args, 'tokens', None):
        data = json.loads(Path(args.tokens).read_text(encoding='utf-8'))
        tokens = list(data.get('tokens', []))
        counts = {
            'files': int(data.get('files', 0)),
            'tags': int(data.get('counts', {}).get('tags', 0)),
            'props': int(data.get('counts', {}).get('props', 0)),
            'components': int(data.get('counts', {}).get('components', 0)),
        }
    else:
        # Choose tokenizer
        use_ast = bool(args.ast) and has_node_tokenizer()
        if args.ast and not has_node_tokenizer():
            print("Warning: --ast requested but Node tokenizer not available; falling back to regex tokenizer.")
        tokens, counts = tokenize_project_ast(args.path) if use_ast else tokenize_project(args.path)

    # Bridge tokens to engine by joining with spaces (engine uses whitespace tokenizer)
    text = " ".join(tokens)

    engine = EmergenceEngine()
    result = engine.process(
        text,
        emergence=args.emergence,
        threshold=args.threshold,
        window=args.window,
        preset=args.emergence_preset,
        mode=args.emergence_mode,
        k=args.k,
        percentile=args.percentile,
        chars=False,
        min_persistence=args.min_persistence,
        hysteresis=args.hysteresis,
        min_gap=args.min_gap,
    )

    # Basic suggestions: list top tokens by frequency (tags/props)
    # Keep simple and useful for now
    freq: dict[str, int] = {}
    for t in tokens:
        freq[t] = freq.get(t, 0) + 1

    def top(prefix: str, n: int = 10) -> list[tuple[str, int]]:
        items = [(k.split(":", 1)[1], v) for k, v in freq.items() if k.startswith(prefix)]
        return sorted(items, key=lambda kv: kv[1], reverse=True)[:n]

    fams = discover_families(args.path)
    suggestions = family_suggestions(fams)

    payload = {
        "input": {
            "path": str(args.path),
            "files": counts["files"],
            "tokens": len(tokens),
            "tags": counts["tags"],
            "props": counts["props"],
            "components": counts["components"],
        },
        "engine": {
            "compression_ratio": result.compression_ratio,
            "mdl_score": result.mdl_score,
            "mdl_grammar_cost": result.mdl_grammar_cost,
            "mdl_data_cost": result.mdl_data_cost,
            "naive_mdl": result.naive_mdl,
            "coverage": result.coverage,
            "valid_lossless": result.valid_lossless,
        },
        "patterns": {
            "rules": result.rules,
            "top_tags": top("TAG:"),
            "top_props": top("PROP:"),
            "top_components": top("IMPORT:"),
        },
        "families": {
            "count": len(fams),
            "suggestions": suggestions,
        },
    }

    if args.emergence:
        payload["emergence"] = {
            "entropies": result.entropies,
            "events": result.events,
            "mdl_trajectory": getattr(result, "mdl_trajectory", None),
        }

    # Exports
    try:
        if args.export_graphviz:
            from emergence_engine.viz import grammar_to_dot
            # Rebuild Grammar from rules
            from emergence_engine.grammar import Grammar
            g = Grammar()
            for lhs, rhs in result.rules.items():
                g.add_rule(lhs, list(rhs))
            args.export_graphviz.write_text(grammar_to_dot(g), encoding="utf-8")
        if args.export_trajectory and args.emergence:
            from emergence_engine.viz import trajectory_to_plotly_json
            if result.entropies and getattr(result, "mdl_trajectory", None):
                import json as _json
                args.export_trajectory.write_text(
                    _json.dumps(trajectory_to_plotly_json(result.entropies, result.mdl_trajectory), indent=2),
                    encoding="utf-8",
                )
    except Exception as e:  # non-fatal
        print(f"Warning: export failed: {e}")

    if args.pretty:
        print(json.dumps(payload, indent=2))
    else:
        print(json.dumps(payload))
    return 0


def build_parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(prog="ufr", description="UFR-DS CLI (Emergence-powered)")
    sub = p.add_subparsers(dest="cmd", required=True)

    a = sub.add_parser("analyze", help="Analyze a React project and report patterns")
    a.add_argument("path", type=Path, help="Path to React project root")
    a.add_argument("--ast", action="store_true", help="Use Babel AST tokenizer (requires Node + deps)")
    a.add_argument("--pretty", action="store_true", help="Pretty-print JSON output")
    a.add_argument("--emergence", action="store_true", help="Compute entropy trajectory and events")
    a.add_argument("--threshold", type=float, default=0.25, help="Emergence normalized curvature threshold")
    a.add_argument("--window", type=int, default=1, help="Emergence detection window (reserved)")
    a.add_argument("--emergence-preset", choices=["sensitive", "balanced", "strict"], help="Preset for emergence detection")
    a.add_argument("--emergence-mode", choices=["static", "adaptive"], default="static", help="Static uses given/preset threshold; adaptive computes from data")
    a.add_argument("--k", type=float, default=3.0, help="Adaptive robust threshold multiplier (MAD)")
    a.add_argument("--percentile", type=float, default=0.9, help="Adaptive percentile (reserved)")
    a.add_argument("--min-persistence", type=int, default=2, help="Emergence: min consecutive steps above threshold")
    a.add_argument("--hysteresis", type=float, default=0.1, help="Emergence: hysteresis margin to end events")
    a.add_argument("--min-gap", type=int, default=2, help="Emergence: minimum gap between events")
    a.add_argument("--tokens", type=Path, help="Path to pre-tokenized JSON (from Node tokenizer)")
    a.set_defaults(func=cmd_analyze)

    # Optional exports for visualization
    a.add_argument("--export-graphviz", type=Path, help="Write final grammar GraphViz DOT to file")
    a.add_argument("--export-trajectory", type=Path, help="Write MDL trajectory + entropies JSON to file (requires --emergence)")

    g = sub.add_parser("generate", help="Generate canonical components from discovered families")
    g.add_argument("path", type=Path, help="Path to React project root (input for discovery)")
    g.add_argument("--out", type=Path, default=Path("generated"), help="Output directory for generated components")
    g.add_argument("--engine", action="store_true", help="Use engine-driven suggestions (analyze first)")
    g.add_argument("--ast", action="store_true", help="Use Babel AST tokenizer when analyzing for engine suggestions")
    g.add_argument("--tokens", type=Path, help="Path to pre-tokenized JSON (from Node tokenizer)")
    g.add_argument("--llm", choices=["openai", "anthropic"], help="Refine generated components with an LLM (optional)")
    g.add_argument("--pretty", action="store_true", help="Pretty-print JSON output")
    g.set_defaults(func=cmd_generate)

    r = sub.add_parser("report", help="Write a Markdown report from analysis JSON")
    r.add_argument("analysis", type=Path, help="Path to JSON payload (from ufr analyze)")
    r.add_argument("--out", type=Path, default=Path("ufr-report.md"), help="Output Markdown file")
    r.add_argument("--llm", choices=["openai", "anthropic"], help="Augment report with LLM-written human-friendly summary")
    r.set_defaults(func=_report_cmd)

    return p


def cmd_generate(args: argparse.Namespace) -> int:
    # Build suggestions
    suggestions: list[dict]
    if args.engine:
        # Obtain tokens for engine analysis if needed
        if args.tokens:
            data = json.loads(Path(args.tokens).read_text(encoding='utf-8'))
            tokens = list(data.get('tokens', []))
        else:
            use_ast = bool(args.ast) and has_node_tokenizer()
            if args.ast and not has_node_tokenizer():
                print("Warning: --ast requested but Node tokenizer not available; falling back to regex tokenizer.")
            tokens, _ = tokenize_project_ast(args.path) if use_ast else tokenize_project(args.path)
        # Run engine to get rules
        engine = EmergenceEngine()
        text = " ".join(tokens)
        result = engine.process(text, emergence=False, chars=False)
        suggestions = suggestions_from_rules(result.rules)
        # Fallback to heuristic if engine found nothing
        if not suggestions:
            fams = discover_families(args.path)
            suggestions = family_suggestions(fams)
    else:
        fams = discover_families(args.path)
        suggestions = family_suggestions(fams)

    files = generate_from_suggestions(suggestions, args.out, llm_provider=args.llm)
    payload = {
        "generated": files,
        "suggestions": suggestions,
    }
    if args.pretty:
        import json
        print(json.dumps(payload, indent=2))
    else:
        import json
        print(json.dumps(payload))
    return 0


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    return args.func(args)


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from .mdl import MDLScorer
from .repair import RePair
from .tokenizer import simple_tokenize, chars as char_tokens
from .emergence import EmergenceDetector


@dataclass
class EngineResult:
    compressed: List[str]
    rules: Dict[str, tuple]
    mdl_score: float
    compression_ratio: float
    mdl_grammar_cost: float
    mdl_data_cost: float
    naive_mdl: float
    coverage: float
    valid_lossless: bool
    entropies: Optional[List[float]] = None
    events: Optional[List[dict]] = None
    mdl_trajectory: Optional[List[Dict[str, float]]] = None
    windows_entropies: Optional[List[float]] = None
    windows_mdl: Optional[List[float]] = None
    window_events: Optional[List[dict]] = None


class EmergenceEngine:
    def __init__(self) -> None:
        self.tokenizer = simple_tokenize
        self.pattern_miner = RePair()
        self.mdl = MDLScorer()

    def process(
        self,
        input_text: str,
        *,
        emergence: bool = False,
        threshold: float = 0.25,
        window: int = 1,
        preset: str | None = None,
        mode: str = "static",
        k: float = 3.0,
        percentile: float = 0.9,
        chars: bool = False,
        min_persistence: int = 2,
        hysteresis: float = 0.1,
        min_gap: int = 2,
    ) -> EngineResult:
        tokens = (char_tokens(input_text) if chars else self.tokenizer(input_text))
        if emergence:
            snapshots = self.pattern_miner.compress_trace(tokens)
            # Use last snapshot as final compressed+grammar
            compressed, grammar = snapshots[-1]
        else:
            compressed, grammar = self.pattern_miner.compress(tokens)
        sigma = len(set(tokens)) or 2
        mdl_components = self.mdl.score_components(grammar, compressed, sigma)
        mdl_score = mdl_components["total"]
        ratio = self.mdl.compression_ratio(tokens, compressed, grammar)

        # Reconstruction and coverage
        reconstructed = self.pattern_miner.reconstruct(compressed, grammar)
        valid_lossless = reconstructed == tokens
        total_tokens = len(reconstructed) if reconstructed else len(tokens)
        covered = 0
        for sym in compressed:
            if sym in grammar.rules:
                # count expansion size as covered
                # Expand one level: sum of terminals in full expansion
                # We can reuse reconstruction of [sym]
                covered += len(self.pattern_miner.reconstruct([sym], grammar))
        coverage = (covered / total_tokens) if total_tokens > 0 else 0.0

        naive = self.mdl.naive_baseline(tokens)
        result = EngineResult(
            compressed=compressed,
            rules={k: tuple(v) for k, v in grammar.as_tuples().items()},
            mdl_score=mdl_score,
            compression_ratio=ratio,
            mdl_grammar_cost=mdl_components["grammar_cost"],
            mdl_data_cost=mdl_components["data_cost"],
            naive_mdl=naive,
            coverage=coverage,
            valid_lossless=valid_lossless,
        )
        if emergence:
            detector = EmergenceDetector(
                threshold=threshold,
                window=window,
                preset=preset,
                mode=mode,
                k=k,
                percentile=percentile,
                min_persistence=min_persistence,
                hysteresis=hysteresis,
                min_gap=min_gap,
            )
            grammars = [g for _, g in snapshots]
            det = detector.detect(grammars)
            # Convert events to plain dicts for JSON friendliness
            result.entropies = det["entropies"]
            # MDL trajectory
            sigma = len(set(tokens)) or 2
            result_mdl = []
            for seq, gram in snapshots:
                comps = self.mdl.score_components(gram, seq, sigma)
                result_mdl.append({"grammar": comps["grammar_cost"], "data": comps["data_cost"], "total": comps["total"]})
            # Grammar differencing for events: compare grammars[i-1] to grammars[i+1]
            def rules_added(prev: Dict[str, tuple], nxt: Dict[str, tuple]) -> list[str]:
                return sorted([lhs for lhs in nxt.keys() if lhs not in prev])

            events_dicts = []
            for e in det["events"]:
                idx = e.index
                if 1 <= idx < len(grammars) - 1:
                    prev = grammars[idx - 1].as_tuples()
                    nxt = grammars[idx + 1].as_tuples()
                    e.rules_added = rules_added(prev, nxt)
                events_dicts.append(
                    {
                        "index": e.index,
                        "magnitude": e.magnitude,
                        "kind": e.kind,
                        "entropy_before": e.entropy_before,
                        "entropy_after": e.entropy_after,
                        "rules_added": e.rules_added,
                    }
                )
            result.events = events_dicts
            result.entropies = det["entropies"]
            # Attach MDL trajectory
            # Using attribute injection since dataclass already defined. Add to dict payload in CLI.
            result.mdl_trajectory = result_mdl
        return result

    def process_sliding(
        self,
        input_text: str,
        *,
        window_size: int,
        step: int,
        threshold: float = 0.25,
        preset: str | None = None,
        mode: str = "static",
        k: float = 3.0,
        percentile: float = 0.9,
        chars: bool = False,
        min_persistence: int = 2,
        hysteresis: float = 0.1,
        min_gap: int = 2,
    ) -> EngineResult:
        tokens = (char_tokens(input_text) if chars else self.tokenizer(input_text))
        # Build token windows
        windows: List[List[str]] = []
        for start in range(0, max(0, len(tokens) - window_size + 1), max(1, step)):
            windows.append(tokens[start : start + window_size])
        if not windows:
            windows = [tokens]

        grammars: List = []
        mdl_totals: List[float] = []
        entropies: List[float] = []

        sigma = len(set(tokens)) or 2
        last_compressed: List[str] = []
        last_grammar = None
        # Process each window independently
        for w in windows:
            compressed, grammar = self.pattern_miner.compress(w)
            comps = self.mdl.score_components(grammar, compressed, sigma)
            mdl_totals.append(comps["total"])
            grammars.append(grammar)
            # Entropy will be computed by detector; we also collect coverage and validity only for last window for summary
            last_compressed, last_grammar = compressed, grammar

        # For summary fields, use last window recon/coverage
        reconstructed = self.pattern_miner.reconstruct(last_compressed, last_grammar) if last_grammar else []
        valid_lossless = reconstructed == (windows[-1] if windows else [])
        total_tokens = len(reconstructed) if reconstructed else len(windows[-1])
        covered = 0
        for sym in (last_compressed or []):
            if last_grammar and sym in last_grammar.rules:
                covered += len(self.pattern_miner.reconstruct([sym], last_grammar))
        coverage = (covered / total_tokens) if total_tokens > 0 else 0.0

        # Naive vs final window MDL
        naive = self.mdl.naive_baseline(windows[-1] if windows else [])

        # Detect events across windows using grammars per window
        detector = EmergenceDetector(
            threshold=threshold,
            window=1,
            preset=preset,
            mode=mode,
            k=k,
            percentile=percentile,
            min_persistence=min_persistence,
            hysteresis=hysteresis,
            min_gap=min_gap,
        )
        det = detector.detect(grammars)

        # Compute rules_added across adjacent windows for event attribution
        events_dicts: List[dict] = []
        for e in det["events"]:
            idx = e.index
            rules_added: List[str] = []
            if 1 <= idx < len(grammars) - 1:
                prev = grammars[idx - 1].as_tuples()
                nxt = grammars[idx + 1].as_tuples()
                rules_added = sorted([lhs for lhs in nxt.keys() if lhs not in prev])
            events_dicts.append(
                {
                    "index": idx,
                    "magnitude": e.magnitude,
                    "kind": e.kind,
                    "entropy_before": e.entropy_before,
                    "entropy_after": e.entropy_after,
                    "rules_added": rules_added,
                }
            )

        # Build a result-like object using final window compressed/grammar for core metrics
        result = EngineResult(
            compressed=last_compressed or [],
            rules={k: tuple(v) for k, v in (last_grammar.as_tuples() if last_grammar else {}).items()},
            mdl_score=mdl_totals[-1] if mdl_totals else 0.0,
            compression_ratio=self.mdl.compression_ratio(windows[-1] if windows else [], last_compressed or [], last_grammar) if last_grammar else 1.0,
            mdl_grammar_cost=0.0,
            mdl_data_cost=0.0,
            naive_mdl=naive,
            coverage=coverage,
            valid_lossless=valid_lossless,
            entropies=det["entropies"],
            events=events_dicts,
            mdl_trajectory=None,
            windows_entropies=det["entropies"],
            windows_mdl=mdl_totals,
            window_events=events_dicts,
        )
        return result



from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, List, Tuple, Optional

from .llm import generate_text


def render_markdown_report(payload: dict) -> str:
    inp = payload.get("input", {})
    eng = payload.get("engine", {})
    pats = payload.get("patterns", {})
    fams = payload.get("families", {})
    lines: List[str] = []
    lines.append(f"# UFR Analysis Report\n")
    lines.append(f"## Input\n")
    lines.append(f"- Path: `{inp.get('path','')}`\n")
    lines.append(f"- Files: {inp.get('files',0)}  Tokens: {inp.get('tokens',0)}\n")
    lines.append(f"- Tags: {inp.get('tags',0)}  Props: {inp.get('props',0)}  Components: {inp.get('components',0)}\n")
    lines.append(f"\n## Engine\n")
    lines.append(f"- Compression ratio: {eng.get('compression_ratio')}\n")
    lines.append(f"- MDL total: {eng.get('mdl_score')} (grammar: {eng.get('mdl_grammar_cost')}, data: {eng.get('mdl_data_cost')})\n")
    lines.append(f"- Coverage: {eng.get('coverage')}  Lossless: {eng.get('valid_lossless')}\n")
    lines.append(f"\n## Top Tags\n")
    for name, count in (pats.get('top_tags') or [])[:10]:
        lines.append(f"- {name}: {count}\n")
    lines.append(f"\n## Top Props\n")
    for name, count in (pats.get('top_props') or [])[:10]:
        lines.append(f"- {name}: {count}\n")
    lines.append(f"\n## Family Suggestions\n")
    for s in (fams.get('suggestions') or [])[:20]:
        lines.append(f"- {s.get('family')}: variants={s.get('suggested_variant_values')} props={s.get('top_props')} members={s.get('members')}\n")
    return "".join(lines)


def render_llm_summary(payload: dict, provider: Optional[str]) -> Optional[str]:
    if not provider:
        return None
    inp = payload.get("input", {})
    eng = payload.get("engine", {})
    pats = payload.get("patterns", {})
    fams = payload.get("families", {})
    # Build a compact structured brief for the model
    brief = {
        "files": inp.get("files"),
        "tokens": inp.get("tokens"),
        "compression_ratio": eng.get("compression_ratio"),
        "coverage": eng.get("coverage"),
        "valid_lossless": eng.get("valid_lossless"),
        "top_tags": (pats.get("top_tags") or [])[:10],
        "top_props": (pats.get("top_props") or [])[:10],
        "suggestions": (fams.get("suggestions") or [])[:10],
    }
    prompt = f"""
You are a senior design systems engineer. Given this analysis, write a concise, human-friendly summary with:
- Key patterns and what they imply (2-4 bullets)
- Suggested consolidations (families and likely variant axes)
- Risks and migration notes (1-2 bullets)
- Clear next actions for the team

Keep it pragmatic and specific to the data. Avoid hype.

Analysis JSON:
{brief}
"""
    return generate_text(prompt, provider)



from typing import Iterable, List


def simple_tokenize(text: str) -> List[str]:
    """Simple whitespace tokenizer.

    - Splits on any whitespace
    - Keeps punctuation as part of tokens
    """
    return [t for t in text.strip().split() if t]


def chars(text: str) -> List[str]:
    """Character-level tokenization (including spaces compressed to \x20)."""
    return [c if c != " " else "\x20" for c in text]




from __future__ import annotations

from typing import Dict, List, Set, Tuple

from .families import base_name


Token = str
RuleMap = Dict[str, Tuple[Token, ...]]


def _expand_rule(rule: str, rules: RuleMap, seen: Set[str] | None = None) -> List[Token]:
    if seen is None:
        seen = set()
    if rule in seen:
        return []
    seen.add(rule)
    out: List[Token] = []
    rhs = rules.get(rule)
    if not rhs:
        return []
    for sym in rhs:
        if sym in rules:
            out.extend(_expand_rule(sym, rules, seen))
        else:
            out.append(sym)
    return out


def _terminals_for_rule(rule: str, rules: RuleMap) -> List[Token]:
    return [t for t in _expand_rule(rule, rules) if t.startswith(("TAG:", "PROP:", "IMPORT:"))]


def suggestions_from_rules(rules: RuleMap) -> List[dict]:
    """Build family suggestions using engine rules (co-occurrence based).

    - Expands rules transitively to terminal tokens
    - Groups TAG:<Name> by base name (Element, Button, etc.)
    - Aggregates co-occurring PROP:* as top props per family
    - Suggests variants from member name differences
    """
    # Collect per-tag props via co-occurrence across rules
    tag_to_props: Dict[str, Dict[str, int]] = {}
    tags_seen: Set[str] = set()
    for rname in list(rules.keys()):
        terminals = _terminals_for_rule(rname, rules)
        tags = [t for t in terminals if t.startswith("TAG:")]
        props = [p for p in terminals if p.startswith("PROP:")]
        for tag in tags:
            tags_seen.add(tag)
            tmap = tag_to_props.setdefault(tag, {})
            for p in props:
                name = p.split(":", 1)[1]
                tmap[name] = tmap.get(name, 0) + 1

    # Group tags into families by base name
    family_members: Dict[str, Set[str]] = {}
    family_props: Dict[str, Dict[str, int]] = {}
    for tag in tags_seen:
        name = tag.split(":", 1)[1]
        base, variant_hints = base_name(name)
        family_members.setdefault(base, set()).add(name)
        # Merge props
        fmap = family_props.setdefault(base, {})
        for p, c in tag_to_props.get(tag, {}).items():
            fmap[p] = fmap.get(p, 0) + c

    # Build suggestions
    suggestions: List[dict] = []
    for base, members in family_members.items():
        if len(members) <= 1:
            continue
        variant_names: Set[str] = set()
        for m in members:
            if m.endswith(base):
                variant_names.add(m[: -len(base)] or "default")
            elif m.startswith(base):
                variant_names.add(m[len(base) :] or "default")
            else:
                variant_names.add(m)
        top_props = sorted(family_props.get(base, {}).items(), key=lambda kv: kv[1], reverse=True)[:10]
        suggestions.append(
            {
                "family": base,
                "members": sorted(members),
                "suggested_variant_values": sorted(v for v in variant_names if v),
                "top_props": [p for p, _ in top_props],
                "source": "engine",
            }
        )
    return suggestions




from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Set, Tuple

from .react_tokenizer import list_source_files, tokenize_file, TAG_OPEN_RE, PROP_RE


COMMON_PREFIXES = [
    "Primary",
    "Secondary",
    "Tertiary",
    "Quaternary",
    "Icon",
    "Fab",
    "Ghost",
    "Outline",
    "Filled",
    "Solid",
    "Link",
]

COMMON_SUFFIXES = [
    "Button",
    "Input",
    "Element",
    "Card",
    "Badge",
    "Chip",
    "Alert",
    "Modal",
    "Dialog",
]


def base_name(name: str) -> Tuple[str, List[str]]:
    """Compute a base name and variant hints from a component-like name.

    Heuristics:
    - If ends with a common suffix (e.g., IconButton, PrimaryButton), base is that suffix.
    - Else remove common prefixes (Primary, Secondary, Icon, etc.) to form base.
    - Return (base, variant_tokens)
    """
    for suf in COMMON_SUFFIXES:
        if name.endswith(suf) and name != suf:
            prefix = name[: -len(suf)]
            variants = [v for v in COMMON_PREFIXES if prefix.startswith(v)]
            variant = prefix if prefix else ""
            return suf, [variant] if variant else variants
    # try strip prefixes
    for pref in COMMON_PREFIXES:
        if name.startswith(pref) and len(name) > len(pref):
            return name[len(pref) :], [pref]
    return name, []


@dataclass
class Element:
    tag: str
    props: List[str] = field(default_factory=list)


def extract_elements_from_text(text: str) -> List[Element]:
    r"""Lightweight extraction of opening JSX tags and their prop names.

    This is a heuristic and may miss complex embedded expressions, but works well
    for common cases. It scans for opening tags and then collects \w-like prop names
    before the next '>' or '/>'.
    """
    elements: List[Element] = []
    for m in TAG_OPEN_RE.finditer(text):
        tag = m.group(1)
        start = m.end(1)
        # Scan ahead to the next closing angle bracket
        end = text.find('>', start)
        if end == -1:
            end = len(text)
        snippet = text[start:end]
        props = [pm.group(1) for pm in PROP_RE.finditer(snippet)]
        elements.append(Element(tag=tag, props=props))
    return elements


def extract_elements(root: Path) -> List[Element]:
    els: List[Element] = []
    for file in list_source_files(root):
        try:
            text = file.read_text(encoding="utf-8", errors="ignore")
        except Exception:
            continue
        els.extend(extract_elements_from_text(text))
    return els


@dataclass
class Family:
    base: str
    members: Set[str] = field(default_factory=set)
    props: Dict[str, int] = field(default_factory=dict)
    variants: Set[str] = field(default_factory=set)

    def add_member(self, name: str, props: List[str], variant_hints: List[str]) -> None:
        self.members.add(name)
        for p in props:
            self.props[p] = self.props.get(p, 0) + 1
        for v in variant_hints:
            if v:
                self.variants.add(v)


def discover_families(root_dir: str | Path) -> Dict[str, Family]:
    root = Path(root_dir)
    els = extract_elements(root)
    families: Dict[str, Family] = {}
    for el in els:
        tag = el.tag
        # Consider only PascalCase tags as components
        if not tag or not tag[0].isupper():
            continue
        b, variant_hints = base_name(tag)
        fam = families.setdefault(b, Family(base=b))
        fam.add_member(tag, el.props, variant_hints)
    return families


def family_suggestions(families: Dict[str, Family]) -> List[dict]:
    """Produce consolidation suggestions for families with multiple members.

    Suggest a canonical component per base with a `variant` prop enumerating member prefixes.
    """
    suggestions: List[dict] = []
    for base, fam in families.items():
        if len(fam.members) <= 1:
            continue
        # Variant names inferred from member names
        variant_names: Set[str] = set()
        for m in fam.members:
            if m == base:
                continue
            # derive likely variant token by removing base suffix/prefix
            if m.endswith(base):
                variant_names.add(m[: -len(base)] or "default")
            elif m.startswith(base):
                variant_names.add(m[len(base) :] or "default")
            else:
                # fallback to full name
                variant_names.add(m)
        if fam.variants:
            variant_names.update(fam.variants)
        # Prop shortlist: top-N props
        top_props = sorted(fam.props.items(), key=lambda kv: kv[1], reverse=True)[:8]
        suggestions.append(
            {
                "family": base,
                "members": sorted(fam.members),
                "suggested_variant_values": sorted(v for v in variant_names if v),
                "top_props": [p for p, _ in top_props],
            }
        )
    return suggestions



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Tuple

from .symbols import Symbol


@dataclass
class ProductionRule:
    lhs: Symbol  # non-terminal
    rhs: List[Symbol]
    frequency: int = 0
    probability: float = 0.0


@dataclass
class Grammar:
    rules: Dict[str, ProductionRule] = field(default_factory=dict)
    terminals: Dict[str, Symbol] = field(default_factory=dict)
    non_terminals: Dict[str, Symbol] = field(default_factory=dict)
    start_symbol: Symbol = field(default_factory=lambda: Symbol("S", kind="nonterminal"))

    def add_rule(self, lhs_name: str, rhs_values: List[str]) -> ProductionRule:
        if lhs_name not in self.non_terminals:
            self.non_terminals[lhs_name] = Symbol(lhs_name, kind="nonterminal")
        lhs = self.non_terminals[lhs_name]
        rhs = [self._get_symbol(v) for v in rhs_values]
        rule = ProductionRule(lhs=lhs, rhs=rhs)
        self.rules[lhs_name] = rule
        return rule

    def _get_symbol(self, value: str) -> Symbol:
        if value in self.non_terminals:
            return self.non_terminals[value]
        if value in self.terminals:
            return self.terminals[value]
        # default new terminals
        sym = Symbol(value, kind="nonterminal" if value.isupper() else "terminal")
        if sym.kind == "terminal":
            self.terminals[value] = sym
        else:
            self.non_terminals[value] = sym
        return sym

    def encode_sequence(self, sequence: List[str]) -> List[str]:
        """Naive encoding: assumes `sequence` already compressed into rule references.

        This method is primarily used by the MDL scorer as a placeholder.
        """
        return list(sequence)

    def as_tuples(self) -> Dict[str, Tuple[str, ...]]:
        return {lhs: tuple(sym.value for sym in rule.rhs) for lhs, rule in self.rules.items()}

    def clone(self) -> "Grammar":
        g = Grammar()
        # Recreate non-terminals/terminals maps based on current symbols
        for name in self.non_terminals:
            g.non_terminals[name] = Symbol(name, kind="nonterminal")
        for name in self.terminals:
            g.terminals[name] = Symbol(name, kind="terminal")
        # Copy rules
        for lhs, rule in self.rules.items():
            g.add_rule(lhs, [sym.value for sym in rule.rhs])
            g.rules[lhs].frequency = rule.frequency
            g.rules[lhs].probability = rule.probability
        return g
